from dataclasses import dataclass, field
from typing import Dict

@dataclass
class Catalog:
    _structure: Dict[str, list[str]] = field(init=False, repr=False)

    def __post_init__(self):
        self.request_base_url = "https://digital.iservices.rte-france.com/open_api"
        self.docs_base_url = "https://data.rte-france.com/catalog/-/api/doc/user-guide"
        self.catalog_base_url = "https://data.rte-france.com/catalog/-/api"
      
        self._meta_mapping = {
            f"{self.request_base_url}/actual_generation/v1/" : {
                "docs_url": f"{self.docs_base_url}/Actual+Generation/1.1",
                "catalog_url": f"{self.catalog_base_url}/generation/Actual-Generation/v1.1",
                "category": "generation",
                "keys": {
                    "actual_generations_per_production_type": {"schema": {"record_path": "values", "meta": ["production_type"]}},
                    "actual_generations_per_unit": {"schema": {"record_path": "values", "meta": [["unit", "eic_code"], ["unit", "name"], ["unit", "production_type"]]}},
                    #"water_reserves": {"schema": {"record_path": "values", "meta": []}}, TBU
                    "generation_mix_15min_time_scale": {"schema": {"record_path": "values", "meta": ["production_type", "production_subtype"]}}
                }
            },
            
            f"{self.request_base_url}/generation_installed_capacities/v1/" : {
                "docs_url": f"{self.docs_base_url}/Generation+Installed+Capacities/1.1",
                "catalog_url": f"{self.catalog_base_url}/generation/Generation-Installed-Capacities/v1.1",
                "category": "generation",
                "keys": {
                    #"capacities_cpc": {"schema": {}},
                    #"capacities_per_production_type": {"schema": {"record_path": "values", "meta": []}}, 
                    "capacities_per_production_unit": {"schema": {"record_path": "values", "meta": [["production_unit", "code_eic"], ["production_unit", "name"], ["production_unit", "location"], ["production_unit", "code_eic_producteur"], ["production_unit", "name_producteur"]]}}
                }
            },
            
            f"{self.request_base_url}/unavailability_additional_information/v6/" : {
                "docs_url": f"{self.docs_base_url}/Unavailability+Additional+Information/6.0",
                "catalog_url": f"{self.catalog_base_url}/generation/Unavailability-Additional-Information/v6.0",
                "category": "generation",
                "keys": {
                    "other_market_information": {"schema": {"meta": ["identifier", "message_id", "version", "creation_date", "unavailability_type", "affected_asset_or_unit_eic_code", "affected_asset_or_unit_name", "affected_asset_or_unit_type", "affected_asset_or_unit_installed_capacity", "reason", "remarks", "status"]}},
                    "transmission_network_unavailabilities": {"schema": {"record_path": ["values_impacted_NTC"], "meta": ["identifier", "message_id", "version", "creation_date", "publication_date", "unavailability_type", "affected_asset_or_unit_eic_code", "affected_asset_or_unit_name", "affected_asset_or_unit_type", "affected_asset_or_unit_installed_capacity", "reason", "remarks", "event_status"]}}, 
                    "generation_unavailabilities": {"schema": {"record_path": ["values"], "meta": ["identifier", "version", "creation_date", "publication_date", "unavailability_type", "fuel_type", "market_participant", "market_participant_eic_code", "affected_asset_or_unit_eic_code", "affected_asset_or_unit_name", "affected_asset_or_unit_type", "affected_asset_or_unit_installed_capacity", "reason", "remarks", "event_status"]}}
                }
            },
            
            f"{self.request_base_url}/generation_forecast/v2/" : {
                "docs_url": f"{self.docs_base_url}/Generation+Forecast/2.1",
                "catalog_url": f"{self.catalog_base_url}/generation/Generation-Forecast/v2.1#",
                "category": "generation",
                "keys": {
                    "forecasts": {"schema": {"record_path": ["values"], "meta": ["type", "production_type", "sub_type"]}}
                },
            },
            
            f"{self.request_base_url}/balancing_energy/v4/" : {
                "docs_url": f"{self.docs_base_url}/Balancing+Energy/4.0",
                "catalog_url": f"{self.catalog_base_url}/market/Balancing-Energy/v4.0",
                "category": "market",
                "keys": {
                    "volumes_per_energy_type": {"schema": {"record_path": ["values"], "meta": ["resolution"]}},
                    "prices": {"schema": {"record_path": ["values"], "meta": ["resolution"]}},
                    "imbalance_data": {"schema": {"record_path": ["values"], "meta": ["resolution"]}},
                    "lead_times": {"schema": {"record_path": ["values"], "meta": ["resolution"]}},
                    "volumes_per_entity_type": {"schema": {"record_path": ["values"], "meta": ["resolution"]}},
                    "tso_offers": {"schema": {"record_path": ["values"], "meta": ["type", "tso_offering", "tso_activating"]}},
                    "volumes_per_reasons": {"schema": {"record_path": ["values"], "meta": ["resolution"]}}
                }
            },
            
            f"{self.request_base_url}/ecowatt/v5/" : {
                "docs_url": f"{self.docs_base_url}/Ecowatt/5.0",
                "catalog_url": f"{self.catalog_base_url}/consumption/Ecowatt/v5.0#",
                "category": "consumption",
                "keys": {
                    "signals": {"schema": {"record_path": ["values"], "meta": ['GenerationFichier', 'jour', 'dvalue', 'message']}}
                }
            },
            
            f"{self.request_base_url}/demand_response/v1/": {
                "docs_url": f"{self.docs_base_url}/231779",
                "catalog_url": f"{self.catalog_base_url}/consumption/Demand-Response/v1.1#",
                "category": "consumption",
                "keys": {
                    "volumes": {"schema": {"meta": ['start_date', 'end_date', 'updated_date', 'eic_code', 'name','trial_nebef_rules_agreement', 'trial_nebef_rules_qualification', 'nebef_rules_recognition', 'nebef_rules_qualification']}}
                }
            },
            
            f"{self.request_base_url}/tempo_like_supply_contract/v1/": {
                "docs_url": f"{self.docs_base_url}/Tempo+Like+Supply+Contract/1.1",
                "catalog_url": f"{self.catalog_base_url}/consumption/Tempo-Like-Supply-Contract/v1.1#",
                "category": "consumption",
                "keys": {
                    "tempo_like_calendars": {"schema": {"record_path": ["values"]}}
                }
            },
            
            f"{self.request_base_url}/consumption/v1/": {
                "docs_url": f"{self.docs_base_url}/Consumption/1.2",
                "catalog_url": f"{self.catalog_base_url}/consumption/Consumption/v1.2#",
                "category": "consumption",
                "keys": {
                    "annual_forecasts": {"schema": {"record_path": ["values"]}},
                    "weekly_forecasts": {"schema": {"record_path": ["values"], "meta": ["updated_date","peak"]}},
                    "short_term": {"schema": {"record_path": ["values"], "meta": ["type"]}}
                }
            },
            
            f"{self.request_base_url}/consolidated_consumption/v1/" : {
                "docs_url": f"{self.docs_base_url}/Consolidated+Consumption/1.0",
                "catalog_url": f"{self.catalog_base_url}/Consolidated-Consumption/v1.0#",
                "category": "consumption",
                "keys": {
                    "consolidated_power_consumption": {"schema": {"record_path": ["values"]}},
                    "consolidated_energy_consumption": {"schema": {"record_path": ["values"]}}
                }
            }
        }
      
        self._requests = {
            key: {
                "request_url": f"{base_url}{key}",
                "docs_url": api["docs_url"],
                "catalog_url": api["catalog_url"],
                "category": api["category"],
                "schema": details["schema"]
                }
                for base_url, api in self._meta_mapping.items()
                for key, details in api["keys"].items()
            }

    @property
    def keys(self) -> str:
        return list(self._requests.keys())
    
    def get_key_content(self, key: str) -> tuple[str]:
        key_content = self._requests.get(key, None)
        if key_content is None:
            raise KeyError(f"Request key '{key}' not in requests catalog")
        request_url = key_content.get("request_url")
        catalog_url = key_content.get("catalog_url")
        docs_url = key_content.get("docs_url", None)
        category = key_content.get("category")
        schema = key_content.get("schema")
        return request_url, catalog_url, docs_url, category, schema
    
    def to_markdown_by_category(self) -> str:
        from collections import defaultdict

        category_tables = defaultdict(list)

        for key in self._requests:
            request_url, catalog_url, docs_url, category, schema = self.get_key_content(key)
            docs_url = docs_url if docs_url is not None else "X"
            row = f"| `{key}` | *[Link]({catalog_url})* | *[Link]({docs_url})* |"
            category_tables[category].append(row)

        sections = []
        for category, rows in category_tables.items():
            section = [
                f"### 📂 {category.capitalize()} Data",
                "",
                "| *data_type* | Catalog URL | Documentation URL |",
                "|-------------------|------------------------|-------------------------|",
                *rows,
                ""  # Ajoute une ligne vide pour la séparation entre les sections
            ]
            sections.append("\n".join(section))

        return "\n\n".join(sections)

    def __repr__(self):
        _repr = "rtedata Catalog : \n"
        for i, key in enumerate(self._requests):
            request_url, catalog_url, docs_url, category, schema = self.get_key_content(key)
            _repr += f"{i} - {key} : \n"
            _repr += f"~> catalog url : {catalog_url} \n"
            if docs_url is not None:
                _repr += f"~> docs url : {docs_url} \n"
        return _repr
        



import requests
from requests.auth import HTTPBasicAuth

from rtedata.tools import Logger
from rtedata.retriever import Retriever
from rtedata.catalog import Catalog

class Client:
  credentials = ["client_id", "client_secret"]
  token_url = "https://digital.iservices.rte-france.com/token/oauth/"
  
  def _get_access_token(self):
    self.logger.info("Generate access token")
    data = {"grant_type": "client_credentials"}
    auth = HTTPBasicAuth(self.client_id, self.client_secret)
    response = requests.post(self.token_url, data=data, auth=auth)

    if response.status_code == 200:
        self.logger.info("Access token generated successfully")
        return response.json()["access_token"]
    else:
        raise Exception(f"Access token error : {response.status_code} - {response.text}")
  
  def retrieve_data(self, start_date: str, end_date: str, data_type: list[str] | str, output_dir: str | None = None):
    if output_dir is None:
      return self.retriever.retrieve(start_date, end_date, data_type)
    else:
      return self.retriever.retrieve(start_date, end_date, data_type, output_dir)

  def __init__(self, client_id: str, client_secret: str):
    self.logger = Logger().logger
    self.catalog = Catalog()

    self.client_id = client_id
    self.client_secret = client_secret
    
    self.token = self._get_access_token()
    self.retriever = Retriever(token=self.token, logger=self.logger, catalog=self.catalog)





import logging

class Logger:
    def __init__(self):
        self.name = "rtedata"
        self.level = logging.INFO
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(self.level)
        
        if not self.logger.handlers:
            formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s')
            ch = logging.StreamHandler()
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)



import os
import time
import requests
import pandas as pd
pd.set_option('future.no_silent_downcasting', True)
import numpy as np

from pathlib import Path
from datetime import datetime, timedelta

from rtedata.tools import Logger
from rtedata.catalog import Catalog

class Retriever:
    def __init__(self, token: str, logger: Logger, catalog: Catalog):
        self.token = token
        self.logger = logger
        self.catalog = catalog

        self.headers = {
            "User-Agent": "rtedata python package (contact: henriupton99@gmail.com)",
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json"
        }

    def _get_request_content_from_key(self, key: str) -> str:
        if key not in self.catalog.keys:
            raise KeyError(f"Invalid input 'data_type' keyword: '{key}'. Must be one of {self.catalog.keys}")
        request_url, catalog_url, docs_url, category, schema = self.catalog.get_key_content(key)
        return request_url, catalog_url, docs_url, category, schema
    
    @staticmethod
    def _convert_date_to_iso8601(date: datetime) -> datetime:
        return date.strftime("%Y-%m-%dT%H:%M:%SZ")

    @staticmethod
    def _convert_date_to_datetime(date_str: str) -> datetime:
        try:
            date = datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")
            return date
        except ValueError:
            raise ValueError("Invalid date format. Expected 'YYYY-MM-DD HH:MM:SS'")
    
    def _generate_tasks(self, start_date: str, end_date: str, base_url: str) -> list[str]:
        task_start_date = start_date
        tasks = []
        while task_start_date < end_date:
            task_end_date = min(task_start_date + timedelta(days=4), end_date)
            start = self._convert_date_to_iso8601(task_start_date - timedelta(days=1))
            end = self._convert_date_to_iso8601(task_end_date + timedelta(days=1))
            tasks.append(f"{base_url}?start_date={start}&end_date={end}")
            task_start_date = task_end_date
        return tasks

    def retrieve(self, start_date: str, end_date: str, data_type: list[str] | str, output_dir: str | None = None) -> dict:
        
        if output_dir is not None:
          output_dir = Path(output_dir)
          if not output_dir.exists():
              os.makedirs(output_dir, exist_ok=True)

        if isinstance(data_type, str):
            data_type = data_type.split(",")

        start_date = self._convert_date_to_datetime(start_date)
        end_date = self._convert_date_to_datetime(end_date)

        if (end_date - start_date).days < 1:
            raise ValueError("Retrieval error : Time difference between input end_date and start_date must be greather than 1 day")
        
        dfs = {}

        for dtype in data_type:
            start_time = time.time()
            request_url, catalog_url, docs_url, category, schema = self._get_request_content_from_key(dtype)
            tasks = self._generate_tasks(start_date, end_date, request_url)
            dfs_tasks = []

            for url in tasks:
                self.logger.info(f"Requesting '{dtype}' from URL: {url}")
                response = requests.get(url, headers=self.headers)

                if response.status_code == 200:
                    data = response.json()
                    data = next(iter(data.values()))
                    df = pd.json_normalize(data, sep="_", errors="ignore", **schema)
                    df = df.replace({None: np.nan})
                    if not df.empty:
                        dfs_tasks.append(df)
                else:
                    self.logger.error(f"Failed to retrieve '{dtype}': {response.status_code} - {response.text}")
                    if docs_url is not None:
                        self.logger.info(f"You can check the related docs at : {docs_url}")
              
            if len(dfs_tasks) == 0:
                self.logger.warning(f"No available data found for data_type={dtype} between given dates. It will then not appear in the resulting dict of datasets")
                if docs_url is not None:
                    self.logger.info(f"You can check the related docs at : {docs_url}")
                continue
            
            for colname in dfs_tasks[0].columns:
                if any(df[colname].isna().all() for df in dfs_tasks):
                    self.logger.warning(f"Detected column '{colname}' with all NA values. It will be removed from resulting dataframe as it is not a consistent column")
                    dfs_tasks = [df.drop([colname], axis=1) for df in dfs_tasks]
            
            df_final = pd.concat(dfs_tasks)
            df_final.drop_duplicates(inplace=True)
                
            if output_dir is not None:
                start = start_date.strftime("%Y%m%d")
                end = end_date.strftime("%Y%m%d")
                filepath = os.path.join(output_dir, f"{dtype}_{start}-{end}.csv")
                df_final.to_csv(filepath, sep=",", index=False)
                self.logger.info(f"Data saved at path : {filepath}")
            
                time.sleep(2)
            
            else:
                elapsed = round(time.time() - start_time, 4)
                self.logger.info(f"Success: '{dtype}' retrieved in {elapsed} seconds")
                dfs[dtype] = df_final

        return dfs



from unittest.mock import patch
from rtedata.client import Client
import pytest

@patch("rtedata.client.requests.post")
def test_get_access_token_success(mock_post):
    mock_post.return_value.status_code = 200
    mock_post.return_value.json.return_value = {"access_token": "FAKE_TOKEN"}

    client = Client(client_id="dummy", client_secret="dummy")
    assert client.token == "FAKE_TOKEN"
    assert mock_post.call_args[0][0] == "https://digital.iservices.rte-france.com/token/oauth/"


@patch("rtedata.client.requests.post")
def test_get_access_token_failure(mock_post):
    mock_post.return_value.status_code = 401
    mock_post.return_value.text = "Unauthorized"

    with pytest.raises(Exception) as e:
        Client(client_id="bad", client_secret="bad")

    assert "Access token error" in str(e.value)



import json
from unittest.mock import patch
from rtedata.retriever import Retriever
from rtedata.catalog import Catalog
from rtedata.tools import Logger

@patch("rtedata.retriever.requests.get")
def test_retrieve_single_data_type(mock_get):
    dummy_token = "FAKE_TOKEN"
    logger = Logger().logger
    catalog = Catalog()

    retriever = Retriever(token=dummy_token, logger=logger, catalog=catalog)

    mock_get.return_value.status_code = 200
    with open("./tests/data/prices.json") as f:
        mock_get.return_value.json.return_value = json.load(f)
    
    result = retriever.retrieve(
        start_date="2024-01-01 00:00:00",
        end_date="2024-01-03 00:00:00",
        data_type="prices"
    )

    assert "prices" in result
    assert not result["prices"].empty

@patch("rtedata.retriever.requests.get")
def test_retrieve_handles_api_error(mock_get):
    dummy_token = "FAKE_TOKEN"
    logger = Logger().logger
    catalog = Catalog()
    retriever = Retriever(token=dummy_token, logger=logger, catalog=catalog)

    mock_get.return_value.status_code = 500
    mock_get.return_value.text = "Server error"

    result = retriever.retrieve(
        start_date="2024-01-01 00:00:00",
        end_date="2024-01-03 00:00:00",
        data_type="prices"
    )

    assert "prices" not in result
